import logging

from .core.fetch_houston2013 import fetch_houston2013
from .core.fetch_houston2013_mmr import fetch_houston2013_mmr
from .core._fetch_houston2013_mat import _fetch_houston2013_mat
from .core.fetch_muufl import fetch_muufl
from .core.fetch_trento import fetch_trento
from .core.fetch_augsburg_ouc import fetch_augsburg_ouc
from .core.fetch_berlin_ouc import fetch_berlin_ouc
from .core.fetch_houston2018_ouc import fetch_houston2018_ouc
from .core._fetch_trento_mat import _fetch_trento_mat
from .core._fetch_muufl_mat import _fetch_muufl_mat
from .core.common import DataMetaInfo

from .util.split_spmatrix import split_spmatrix
from .util.fileio import read_roi
from .util.hsi2rgb import hsi2rgb


__all__ = [
    'fetch_houston2013', 
    'fetch_houston2013_mmr', 
    'fetch_muufl', 
    'fetch_trento', 
    'fetch_berlin_ouc',
    'fetch_augsburg_ouc', 
    'fetch_houston2018_ouc',
    '_fetch_trento_mat',
    '_fetch_muufl_mat',
    '_fetch_houston2013_mat',
    'split_spmatrix', 
    'read_roi', 
    'hsi2rgb',
    'DataMetaInfo',
]


# If torch is imported, add these API
try:
    import torch
except ImportError:
    pass
else:
    from .torch.datasets import Houston2013, Muufl, Trento, Houston2013Mmr, _Houston2013Mmrs, BerlinOuc, AugsburgOuc, Houston2018Ouc, _MuuflMat, _TrentoMat, _Houston2013Mat
    from .torch.common_hsi_dsm_dataset import CommonHsiDsmDataset
    from .util.lbl2rgb import lbl2rgb
    from .util.classification_mapper import ClassificationMapper

    __all__ += [
        'CommonHsiDsmDataset', 
        'Houston2013',
        'Houston2013Mmr', 
        'Muufl', 
        'Trento', 
        'BerlinOuc', 
        'AugsburgOuc',
        'Houston2018Ouc',
        '_MuuflMat', 
        '_TrentoMat', 
        '_Houston2013Mmrs',
        '_Houston2013Mat',
        'lbl2rgb',
        'ClassificationMapper'
    ]








# SPDX-FileCopyrightText: 2025-present songyz2023 <songyz2023dlut@outlook.com>
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
import unittest

import numpy as np
import skimage
from rs_fusion_datasets import fetch_houston2013_mmr, _Houston2013Mmrs, fetch_houston2018
from torch.utils.data import DataLoader
from itertools import product
from hsi2rgb import hsi2rgb
import torch
from typing import get_type_hints, TypedDict, get_origin

from rs_fusion_datasets import CommonHsiDsmDataset, DataMetaInfo
from rs_fusion_datasets.util.fileio import ftp_download


def is_typeddict_instance(obj, typeddict_cls):
    if not isinstance(obj, dict):
        return False
    type_hints = get_type_hints(typeddict_cls)
    for key, expected_type in type_hints.items():
        if key not in obj: # do not check value: isinstance() argument 2 cannot be a parameterized generic
            print(f"Key '{key}' is missing or has incorrect type.")
            return False
    return True

    

class Test(unittest.TestCase):
    def setUp(self):
        pass

    def torch_dataloader_test(self, dataset :CommonHsiDsmDataset):
        b = 16
        dataloader = DataLoader(dataset, batch_size=b, shuffle=True, drop_last=True)
        x_h, x_l, y, extras = next(iter(dataloader))
        
        n_test = 10
        for x_h, x_l, y, extras in dataloader:
            if torch.cuda.is_available():
                x_h, x_l, y = x_h.cuda(), x_l.cuda(), y.cuda()
            self.assertIsInstance(x_h, torch.Tensor)
            self.assertIsInstance(x_l, torch.Tensor)
            self.assertIsInstance(y, torch.Tensor)
            self.assertEqual(x_h.shape, torch.Size([b, dataset.INFO['n_channel_hsi'],dataset.patch_size,dataset.patch_size]))
            self.assertEqual(x_l.shape, torch.Size([b, 1,dataset.patch_size,dataset.patch_size]))
            self.assertEqual(y.shape, torch.Size([b, dataset.n_class]))
            self.assertEqual(x_h.dtype, torch.float)
            self.assertEqual(x_l.dtype, torch.float)
            self.assertEqual(y.dtype, torch.float)
            if n_test <= 0:
                break
            else:
                n_test -= 1

    def generate_lbl2rgb(self, truth, info, subset):
        h,w = truth.shape
        y = np.eye(info['n_class']+1)[truth.todense()].transpose(2, 0, 1) # One Hot
        self.assertEqual(y.shape, (info['n_class']+1, h, w))

        rgb = lbl2rgb(y, info['name'])
        self.assertEqual(rgb.shape, (3, h, w))
        self.assertLessEqual(rgb.max(), 1.0)
        self.assertGreaterEqual(rgb.min(), 0.0)

        img = (rgb*255.0).astype(np.uint8).transpose(1, 2, 0)
        skimage.io.imsave(f"dist/{info['name']}_{subset}.png", img, check_contrast=False)

    def test_fetch_houston2013(self):
        casi, lidar, train_truth, test_truth, info = fetch_houston2013_mmr()
        H, W = 349, 1905
        C_H, C_L = 144, 1
        self.assertEqual(train_truth.data.max(), info['n_class'])
        self.assertEqual(train_truth.data.min(), 1)
        self.assertEqual(train_truth.todense().min(), 0)
        self.assertEqual(test_truth.data.max(), info['n_class'])
        self.assertEqual(test_truth.data.min(), 1)
        self.assertEqual(test_truth.todense().min(), 0)
        self.assertEqual(casi.shape, (C_H, H, W))
        self.assertEqual(lidar.shape, (C_L, H, W))
        self.assertEqual(train_truth.shape, (H, W))
        self.assertEqual(test_truth.shape, (H, W))
        self.assertEqual(info['n_channel_hsi'], 144)
        self.assertEqual(info['n_channel_lidar'], 1)
        self.assertEqual(info['n_class'], 15)
        self.assertEqual(info['width'], W)
        self.assertEqual(info['height'], H)
        self.assertEqual(len(info['label_name']), 15)
        self.assertEqual(info['wavelength'].shape, (144,))
        self.assertEqual(info['wavelength'][0], 364.000000)
        self.assertEqual(info['wavelength'][-1], 1046.099976)
        self.assertEqual(len(info['wavelength']), C_H)
        self.assertTrue(is_typeddict_instance(info, DataMetaInfo))

        hsi = casi.astype(np.float32)
        hsi = (hsi - hsi.min()) / (hsi.max() - hsi.min())
        rgb = hsi2rgb(hsi, wavelength=info['wavelength'], input_format='CHW', output_format='HWC')
        skimage.io.imsave(f"dist/{info['name']}_hsi.png", (rgb * 255.0).astype(np.uint8))

        dsm = lidar[0, :, :]
        dsm_img = (dsm - dsm.min()) / (dsm.max() - dsm.min()) * 255.0
        skimage.io.imsave(f"dist/{info['name']}_dsm.png", dsm_img.astype(np.uint8))

    def test_torch_datasets(self):
        for Dataset, subset, patch_size in product([_Houston2013Mmrs], ['train', 'test', 'full'], [1, 5, 10, 11]):
            dataset = Dataset(subset=subset, patch_size=patch_size)
            self.torch_dataloader_test(dataset)

    def test_fetch_houston2018(self):
        hsi, lidar, lbl, info = fetch_houston2018()
        hsi = hsi.astype(np.float32)
        hsi = (hsi - hsi.min()) / (hsi.max() - hsi.min())
        rgb = hsi2rgb(hsi, wavelength=info['wavelength'], input_format='CHW', output_format='HWC')
        skimage.io.imsave(f"dist/{info['name']}_hsi.png", (rgb * 255.0).astype(np.uint8))

        dsm = lidar[0, :, :]
        dsm_img = (dsm - dsm.min()) / (dsm.max() - dsm.min()) * 255.0
        skimage.io.imsave(f"dist/{info['name']}_dsm.png", dsm_img.astype(np.uint8))


if __name__ == '__main__':
    unittest.main()


from typing import List, Tuple, Union, Optional
from pathlib import Path
import numpy as np
from numpy import ndarray
import scipy.io

from scipy.sparse import coo_array
from jaxtyping import  UInt16, Float64, UInt8, Int8, Int4

from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo


def fetch_berlin_ouc(
        url       :Union[str, List[str]]      = 'https://github.com/songyz2019/rs-fusion-datasets-dist/releases/download/v1.0.0/berlin-ouc.zip', 
        data_home :Optional[Union[Path, str]] = None
    ) -> Tuple[
        UInt16[ndarray, '180 1723 476'], 
        Float64[ndarray, '4 1723 476'],
        UInt8[coo_array, '1723 476'],
        UInt8[coo_array, '1723 476'],
        UInt8[coo_array, '1723 476'],
        DataMetaInfo
    ]:
    """Fetch and load the Berlin (ouc) dataset. 

    The background is 0, and the labels start from 1. All images are CHW formats.
    
    :param url: The URL to download the dataset. Use a list to specify multiple mirrored URLs.
    :param data_home: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    :return: (hsi, lidar, train_labels, train_labels, info)
    """

    basedir = zip_download_and_extract('berlin-ouc', url, {
        'berlin-ouc.zip'         : '5fdd156b0283bddab3cf9916f34d3d6b3a2ceb68e7c38c3f8022292199aafcf8',
        'Berlin/berlin_gt.mat'   : 'dd9286cd565535857721627878de3cbebffa2dacc625b539474d16cba3a7343e',
        'Berlin/berlin_hsi.mat'  : '580f397d9f249b4ec3f02b8cc570d910d765afa9f77b17d8a4f4799edee94b1f',
        'Berlin/berlin_index.mat': '1a66ec454ea95bbf5782525d9fcf9a04412a948b7b3aa6919ed324516d42a8b5',
        'Berlin/berlin_sar.mat'  : 'a6cf2f86db7f1a02c340e813a7b46f9c6c58e297e1eb9b5fa19ed49244b2b8a7'
    }, data_home=data_home)


    # 3. 数据加载
    hsi   :Float64[ndarray, '1723 476 244']= scipy.io.loadmat(
        basedir / 'Berlin/berlin_hsi.mat',
    )['berlin_hsi']

    sar :Float64[ndarray, '1723 476 4']   = scipy.io.loadmat(
        basedir / 'Berlin/berlin_sar.mat',
    )['berlin_sar']

    lbl_map :UInt8[ndarray, '1723 476']   = scipy.io.loadmat(
        basedir / 'Berlin/berlin_gt.mat',
    )['berlin_gt']
    lbl_subsets = scipy.io.loadmat(basedir / 'Berlin/berlin_index.mat')
    lbl_train_coord :Int8[ndarray, 'M 2']    = lbl_subsets['berlin_train']   # TODO: replace M N Q with the actual number
    lbl_test_coord  :Int8[ndarray, 'N 2']  = lbl_subsets['berlin_test']
    lbl_all_coord   :Int4[ndarray, 'Q 2'] = lbl_subsets['berlin_all']
    lbl_train :UInt8[coo_array, '1723 476'] = coo_array((
        lbl_map[ lbl_train_coord[:,0],lbl_train_coord[:,1] ],
        lbl_train_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_test :UInt8[coo_array, '1723 476']  = coo_array((
        lbl_map[ lbl_test_coord[:,0],lbl_test_coord[:,1] ],
        lbl_test_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_all  :UInt8[coo_array, '1723 476']  = coo_array((
        lbl_map[ lbl_all_coord[:,0],lbl_all_coord[:,1] ],
        lbl_all_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)

    info :DataMetaInfo = {
        'name': 'berlin-ouc',
        'full_name': 'preprocessed berlin dataset',
        'version': '',
        'homepage': 'https://github.com/oucailab/DCMNet/',
        'license': '',
        'n_channel_hsi': 244,
        'n_channel_dsm': 4,
        'n_class': 9,
        'width': 1723,
        'height': 476,
        'label_name': {
            1: 'Forest',
            2: 'Residential',
            3: 'Industrial',
            4: 'Lowplants',
            5: 'Soil',
            6: 'Allotment',
            7: 'Commercial',
            8: 'Water'
        },
        'wavelength': np.linspace(400, 2500, 244) # TODO: check the wavelength
    }

    return hsi.transpose(2,0,1), sar.transpose(2,0,1), lbl_train, lbl_test, lbl_all, info



__all__ = ['fetch_trento']



from typing import Literal
import numpy as np

__all__ = ['hsi2rgb']

def _cie_xyz_weights(n):
    """

    n 波长
    """

    def g(x, mu, sigma1, sigma2):
        """分段高斯函数"""
        return np.where(
            x < mu,
            np.exp(-(x - mu) ** 2 / (sigma1 ** 2) / 2),
            np.exp(-(x - mu) ** 2 / (sigma2 ** 2) / 2),
        )
    x = (
        0.362   * g(n, 442.0 , 16.0 , 26.7)
        -0.065  * g(n, 501.1, 20.4 , 26.2 )
        + 1.056 * g(n, 599.8, 37.9 , 31.0 )
    )
    y = (
        0.286   * g(n, 530.9 , 16.3 , 31.1)
        + 0.821 * g(n, 568.8, 46.9 , 40.5 )
    )
    z = (
        0.980   * g(n, 437.0 , 11.8 , 36.0)
        + 0.681 * g(n, 459.0, 26.0 , 13.8 )
    )
    return x,y,z

# This should be a clear, and straightforward core functions, without any unnecessary complexity, since the algorithm itself is complex enough.
def _hsi2rgb(hsi, wavelength):
    """
    Convert an HSI (Hyperspectral Imaging) Image to RGB format. This is a raw fuctions without any checking and convertions of inputs and outputs.
    
    Parameters:
    hsi : numpy.ndarray
        The input HSI data, expected to be a 3D array of shape (height, width, spectral bands) where the last dimension represents different spectral bands.
        The values are expected to be in the range [0, 1].
    wavelength : numpy.ndarray
        The wavelengths corresponding to the spectral bands in the HSI data. Expected to be a 1D array of the same length as the number of spectral bands.
        The values are expected to be in nanometers.
    raw : bool, optional
        If True, the function will return the raw RGB values without rescaling intensity. Default is False.
    
    Returns:
    numpy.ndarray
        The RGB representation of the input HSI data, with shape (height, width, 3). The values are rescaled to the range [0, 1] if raw is False.
    """
    
    x,y,z = _cie_xyz_weights(wavelength)
    X = np.sum(hsi*x, axis=-1)
    Y = np.sum(hsi*y, axis=-1)
    Z = np.sum(hsi*z, axis=-1)

    transform_matrix = [
        [3.1338561, -1.6168667, -0.4906146],
        [-0.9787684, 1.9161415, 0.0334540],
        [0.0719453, -0.2289914, 1.4052427]
    ]

    xyz = np.stack([X, Y, Z], axis=-1)
    rgb = np.einsum('C c, ... h w c -> ... h w C', transform_matrix, xyz)

    return rgb


# This should be a wrapper for _hsi2rgb, make it user-friendly and compatible with many inputs.
def hsi2rgb(hsi, wavelength_range=(350,1000), wavelength=None, input_format: Literal['CHW', 'HWC'] = 'CHW', output_format: Literal['CHW', 'HWC', 'Same'] = 'Same', xyz_to_rgb=True, gamma=2.2):
    """
    Converts a hyperspectral image (HSI) to an RGB image. 
    Parameters:
        hsi (numpy.ndarray, torch.Tensor, or jax.numpy.ndarray): 
            The input hyperspectral image. It can be a numpy array, a PyTorch tensor, or a JAX array.
        wavelength_range (tuple, optional): 
            The range of wavelengths (in nanometers) corresponding to the hyperspectral bands. 
            Defaults to (350, 1000) which looks fine for most sensors.
        wavelength (numpy.ndarray, optional): 
            In most case, you just need wavelength_range, and leave wavelength None. A 1D array specifying the exact wavelengths for each band. If None, it will be 
            generated as a linear space within `wavelength_range`. Defaults to None.
        input_format (Literal['CHW', 'HWC'], optional): 
            The format of the input HSI. 'CHW' indicates (Channels, Height, Width), and 'HWC' 
            indicates (Height, Width, Channels). Defaults to 'CHW'.
        output_format (Literal['CHW', 'HWC', 'Same'], optional): 
            The desired format of the output RGB image. 'Same' will use the same format as 
            `input_format`. Defaults to 'Same'.
        xyz_to_rgb (bool, optional): 
            Whether to normalize and apply gamma correction to the RGB output. Defaults to True.
        gamma (float, optional): 
            The gamma correction factor to apply if `xyz_to_rgb` is True. Defaults to 2.2.
    Returns:
        numpy.ndarray, torch.Tensor, or jax.numpy.ndarray: 
            The converted RGB image in the specified `output_format`. The type of the output 
            matches the type of the input HSI.
    Raises:
        ValueError: 
            If the input HSI is not a numpy array, PyTorch tensor, or JAX array.
        ValueError: 
            If `input_format` or `output_format` is not one of the allowed values.
    Notes:
        - The function automatically normalizes the input HSI to the range [0, 1] if its values 
          are outside this range.
        - The RGB output is always normalized to the range [0, 1] before gamma correction.
        - The function supports batch dimensions, which are preserved in the output.
        - These doc is AI generated and audited by human
    """
    if output_format == 'Same':
        output_format = input_format

    tensor_type = 'numpy'
    if not isinstance(hsi, np.ndarray):
        # Guess the type of hsi and convert to numpy array
        try:
            import torch
            # Note: This is a little trick beacause we don't add torch or jax as a dependency, it will make this package very large.
            if isinstance(hsi, torch.Tensor):
                hsi = hsi.cpu().detach().numpy()
                tensor_type = 'torch'
        except ImportError:
            pass
    if not isinstance(hsi, np.ndarray):    
        try:
            import jax
            if isinstance(hsi, jax.numpy.ndarray):
                hsi = np.array(hsi)
                tensor_type = 'jax'
        except ImportError:
            pass
    if not isinstance(hsi, np.ndarray): 
        raise ValueError("hsi should be a numpy array, a torch tensor or a jax array")

    b=range(hsi.ndim - 3) # the batch dimension**s**
    # Convert any input shape to HWC
    if input_format == 'CHW':
        hsi = np.transpose(hsi, (*b, -2, -1, -3))
    elif input_format == 'HWC':
        pass
    else:
        raise ValueError("format should be 'HWC' or 'CHW'")
    
    if hsi.max() > 1 or hsi.min() < 0:
        max_hsi = np.max(hsi, axis=(-2, -3), keepdims=True)
        min_hsi = np.min(hsi, axis=(-2, -3), keepdims=True)
        hsi = (hsi - min_hsi)/(max_hsi-min_hsi)
    # --------- HWC Begin -----------
    if wavelength is None:
        wavelength = np.linspace(wavelength_range[0], wavelength_range[-1], hsi.shape[-1])
        # Use wavelength_range[-1] instead of wavelength_range[1] in case users pass the wrong parameter.
    rgb = _hsi2rgb(hsi, wavelength)
    if xyz_to_rgb:
        max_rgb = np.max(rgb, axis=(-2, -3), keepdims=True)
        min_rgb = np.min(rgb, axis=(-2, -3), keepdims=True)
        rgb = (rgb - min_rgb)/(max_rgb-min_rgb)
        rgb = np.power(rgb/float(np.max(rgb)), 1/gamma) if gamma != 1 else rgb
    # --------- HWC End -----------

    # Note that rgb is always hwc here
    if output_format=='CHW':
        rgb = np.transpose(rgb, (*b, -1, -3, -2))

    if tensor_type == 'torch':
        import torch
        rgb = torch.from_numpy(rgb)
    elif tensor_type == 'jax':
        import jax
        rgb = jax.device_put(rgb)
    return rgb



from ftplib import FTP
import logging
import os
from pathlib import Path
import re
from typing import List, Union, Optional, Tuple, Dict
import warnings
import hashlib
from io import StringIO
from zipfile import ZipFile

import numpy as np
import scipy
from scipy.sparse import coo_array, sparray
from jaxtyping import UInt16
import urllib


def get_data_home(data_home :Optional[Union[Path, str]]=None) -> Path:
    if data_home is None:
        data_home = os.environ.get("SCIKIT_LEARN_DATA", os.path.join("~", "scikit_learn_data")) # Not neat but works
        data_home = Path(data_home)
    elif isinstance(data_home, str):
        data_home = Path(data_home)
    data_home = data_home.expanduser()
    data_home.mkdir(parents=True, exist_ok=True)
    return data_home


def read_roi(path :Path, shape :Tuple[int, int]) -> UInt16[sparray, 'h w']:
    """
    读取ENVI软件导出roi文件得到的txt文件,得到一个稀疏矩阵图像

    用起来像字典

    :param path: 文件路径
    :return: An coo_array image representing the ROI
    """
    warnings.simplefilter("ignore", category=UserWarning) # Supress loadtxt's warning when data is empty

    img = coo_array(shape, dtype='uint16')
    buf = ""
    cid = 1
    
    with open(path, 'r') as f:
        for line in f:
            # Comments
            if line.startswith(";") or line.isspace():
                continue

            # Seprator
            if line.lstrip().startswith("1 "): # magick string for compatibility
                data=np.loadtxt(StringIO(buf), usecols=(2, 1), comments=';', dtype='uint16', converters=float)
                buf = ""
                if data.size > 0:
                    rows,cols = data.T
                    vals = cid*np.ones_like(rows)
                    cid += 1
                    # breakpoint()
                    img += coo_array((vals,(rows,cols)), shape=shape) # There may be duplicate points in roi file, so these steps are essential
                    img.data[img.data>cid] = 0  # 清除重复像素点
            # Data
            else:
                buf += line

        # Read last block  
        if buf!="":
            data=np.loadtxt(StringIO(buf), usecols=(2, 1), comments=';', dtype='uint16')
            buf = ""
            if data.size > 0:
                rows,cols = data.T
                vals = cid*np.ones_like(rows)
                img += coo_array((vals,(rows,cols)), shape=shape)


    warnings.resetwarnings()

    return img.tocoo()

def verify_file(path: Path, expected_sha256: str) -> bool:
    def _quick_sha256(path):
        """Calculate the sha256 hash of the file at path."""
        sha256hash = hashlib.sha256()
        chunk_size = 8192
        with open(path, "rb") as f:
            while True:
                buffer = f.read(chunk_size)
                if not buffer:
                    break
                sha256hash.update(buffer)
        return sha256hash.hexdigest()

    logging.debug(f"Verifying {path} with sha256 {expected_sha256}")
    if not path.exists():
        logging.debug(f"Not exist")
        return False
    if not path.is_file():
        logging.debug(f"Not a file")
        raise ValueError(f"{path} is not a file")
    if expected_sha256 != _quick_sha256(path):
        logging.debug(f"Incorrect sha256: {_quick_sha256(path)}")
        return False
    logging.debug(f"PASS")
    return True

def zip_download_and_extract(dir_name :str, url :Union[str, List[str]], required_files :Dict[str, str], data_home :Optional[Union[str,Path]] = None) -> None:
    """
    Download a zip file from url and extract it to datahome

    By calling this, either of the two scenarios will be satisfied:
    1. The program raise an Exception
    2. All files in required_files are found and verified in datahome
    The required_files[dir_name+'.zip'] is the zip file itself, and the rest are the files in the zip file.
    Put your file to basedir/zip_name to skip the download step.
    """
    # breakpoint()
    dir_name = dir_name.removesuffix('/')
    if isinstance(dir_name, str):
        basedir = Path(get_data_home(data_home))/dir_name
        basedir.mkdir(parents=True, exist_ok=True)
    else:
        raise ValueError(f"dir_name is not a valid string or Path: {dir_name}")
        
    
    # Everything is ok, return
    if all([verify_file(basedir/path, sha256) for path,sha256 in required_files.items()]):
        return basedir
    

    zip_name = dir_name+'.zip'
    mirrored_download(
        basedir/zip_name,
        url,
        required_files[zip_name]
    )

    # Now the zip file must exist and verified, or the program raised an exception
    # Assuming the zip file is OK

    with ZipFile(basedir/zip_name, 'r') as f_zip:
            f_zip.extractall(basedir)

    for path, sha256 in required_files.items():
        if path == zip_name:
            continue
        if not verify_file(basedir/path, sha256):
            raise ValueError(f"{basedir/path} is not found or not verified. If it exists, please delete {basedir} and try again.")

    return basedir

def _ftp_download(path :Path, url, sha256: str) -> None:
    """Only simple case url supported ftp://username@host.com?password/aug.zip"""
    grp = re.match(r'ftp://(\w+)@([\w\.]+)\?([\w\.]+)/([\w\.]+)', url)
    ftp_username, ftp_host, ftp_password, ftp_file_name = grp.groups()

    if verify_file(path, sha256):
        return
    path.parent.mkdir(parents=True, exist_ok=True)

    ftp = FTP(ftp_host)
    ftp.login(user=ftp_username, passwd=ftp_password)
    ftp.voidcmd('TYPE I')

    size_mb = ftp.size(ftp_file_name) / 1024 / 1024
    with open(path, 'wb') as fp:
        global i_rsfds
        i_rsfds = 0 # add a random suffix
        def retr_callback(chunk: bytes):
            global i_rsfds
            i_rsfds += len(chunk)
            print(f"{i_rsfds/1024/1024:.3f}MB/{size_mb:.3f}MB", end='\r')
            fp.write(chunk)
        ftp.retrbinary(f'RETR {ftp_file_name}', retr_callback, blocksize=65536)

    assert verify_file(path, sha256)





def mirrored_download(path :Path, url :Union[str, List[str]], sha256: str) -> None:
    """Download a file from urls, and check the sha256 hash. If the download fails, try the next url.

    :param path: The path to save the file
    :param urls: The list of urls to download from
    :param sha256: The sha256 hash of the file
    """
    if verify_file(path, sha256):
        return
    path.parent.mkdir(parents=True, exist_ok=True)
    if isinstance(url, str):
        url = [url,]
    
    opener = urllib.request.build_opener()
    opener.addheaders = [('User-Agent', 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36 Edg/135.0.0.0')]
    urllib.request.install_opener(opener)

    success = False
    for url in url:
        try:
            print(f"Downloading {url} to {path}")
            if url.startswith('http://') or url.startswith('https://'):
                urllib.request.urlretrieve(url, path, reporthook=lambda blocknum, bs, size: print(f"{blocknum*bs/1024/1024:0.3f}MB/{size/1024/1024:0.3f}MB", end="\r")  if blocknum%128==0 else None )
            elif url.startswith('ftp://'):
                _ftp_download(path, url, sha256)
            print("\nDownload Success")
            if verify_file(path, sha256):
                success = True
                break
        except (urllib.error.URLError, urllib.error.HTTPError, ValueError) as e:
            print(f"Error downloading {url}: {e}")
            continue

    urllib.request.install_opener(urllib.request.build_opener()) # Reset
    if not success:
        raise ValueError(f"Failed to download {path} from all urls")
    else:
        return


def load_one_key_mat(path :Path, *args, **kwargs):
    mat = scipy.io.loadmat(path, squeeze_me=True, mat_dtype=True, struct_as_record=False, *args, **kwargs)
    keys = [x for x in mat.keys() if not x.startswith('__') and not x.endswith('__')]
    assert len(keys) == 1, f"Mat file {path} has more than one key: {keys}, please load it manually"
    return mat[keys[0]]



from typing import Tuple
import numpy as np
from scipy.sparse import coo_array
from contextlib import contextmanager

@contextmanager
def fixed_seed_rng(seed):
    rng = np.random.default_rng(seed)
    yield rng

def split_spmatrix(a :coo_array, n_sample_perclass=100, seed=0x0d00) -> Tuple[coo_array, coo_array]:
    """
    Split a sparse matrix into train and test sets.
    The train set contains `n_sample_perclass` samples (if n_sample_perclass is int) or percentage of a class (if n_sample_perclass is float) for each class, and the test set contains the rest.
    """
    with fixed_seed_rng(seed) as rng:
        train = coo_array(([],([],[])),a.shape, dtype='int')
        n_class = a.data.max()
        for cid in range(1,n_class+1):
            N = len(a.data[a.data==cid])
            if isinstance(n_sample_perclass, float) and n_sample_perclass < 1.0:
                n_sample_perclass = int(N * n_sample_perclass)
            indice = rng.choice(N, n_sample_perclass, replace=False)
            row = a.row[a.data==cid][indice]
            col = a.col[a.data==cid][indice]
            val = np.ones(len(row)) * cid
            train += coo_array((val, (row, col)), shape=a.shape, dtype='int')
        test = (a - train)
    return train.tocoo(),test.tocoo()



from typing import Literal,Tuple, Dict

from scipy.sparse import coo_array
from torchvision.datasets.vision import VisionDataset
import torch
import numpy as np
from numpy import ndarray
from jaxtyping import Num, Float

from ..core.common import DataMetaInfo

def _channel_wise_normalize(x: Float[ndarray, 'c h w']) -> Float[ndarray, 'c h w']:
    """
    Normalize each channel of the input image to the range [0, 1].
    :param x: Input image with shape (c, h, w).
    :return: Normalized image with shape (c, h, w).
    """
    x = x.astype(np.float32)
    min_val = np.min(x, axis=(-1, -2), keepdims=True)
    max_val = np.max(x, axis=(-1, -2), keepdims=True)
    return (x - min_val) / (max_val - min_val)  # Avoid division by zero

class CommonHsiDsmDataset(VisionDataset):
    def __init__(self,
                 hsi :Num[ndarray, 'c h w'], 
                 dsm :Num[ndarray, 'd h w'],
                 lbl_train :Num[ndarray, 'h w'],
                 lbl_test :Num[ndarray, 'h w'],
                 info: DataMetaInfo,
                 split: Literal['train', 'test', 'full'], 
                 patch_size: int = 5,  # I prefer patch_radius, but patch_size is more popular and maintance two patch_xxx is too complex...
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load truth
        self.subset = split
        if self.subset == 'train':
            self.lbl = lbl_train
        elif self.subset == 'test':
            self.lbl = lbl_test
        elif self.subset == 'full':
            self.lbl = coo_array(-1*np.ones_like(lbl_train.todense(), dtype=np.int16), dtype='int')
        else:
            raise ValueError(f"Unknown subset: {split}")

        # Load patch size
        if patch_size % 2 != 1:
            pass
            # warnings.warn("Non-odd patch size may cause unknown behaviors (classification pixel is at right bottom 2x2 center location). Use at your own risk.",category=UserWarning,stacklevel=2)
        self.patch_size = patch_size
        self.patch_radius = patch_size // 2 # if patch_size is odd, it should be (patch_size - w_center)/2, but some user will use on-odd patch size

        # Load basic info
        self.HSI, self.DSM, self.INFO = hsi, dsm, info
        self.n_class = self.INFO['n_class']
        self.n_channel_hsi = self.INFO['n_channel_hsi']
        self.n_channel_dsm = self.INFO['n_channel_dsm']

        # Preprocess HSI
        pad_shape = ((0, 0), (self.patch_radius, self.patch_radius), (self.patch_radius, self.patch_radius))
        self.hsi = _channel_wise_normalize(self.HSI)
        self.hsi = np.pad(self.hsi, pad_shape, mode='reflect')
        self.hsi = torch.from_numpy(self.hsi).float()

        # Preprocess DSM
        self.dsm = _channel_wise_normalize(self.DSM)
        self.dsm = np.pad(self.dsm, pad_shape, mode='reflect')
        self.dsm = torch.from_numpy(self.dsm).float()


    def __len__(self):
        return self.lbl.count_nonzero()

    def __getitem__(self, index) -> Tuple[
            Float[ndarray, 'c h w'],
            Float[ndarray, 'c h w'],
            Float[ndarray, 'c h w'],
            Dict[str, int]
        ]:
        w = self.patch_size

        i = self.lbl.row[index]
        j = self.lbl.col[index]
        cid = self.lbl.data[index].item()

        x_hsi = self.hsi[:, i:i+w, j:j+w]
        x_dsm = self.dsm[:, i:i+w, j:j+w]
        y = np.eye(self.n_class)[cid-1]
        y = torch.from_numpy(y).float()

        # 额外信息: 当前点的坐标
        extras = {
            "location": [i, j],
            "index": index
        }

        return x_hsi, x_dsm, y, extras

    @property
    def uid(self) -> str:
        '''a uid for logging and batch training'''
        return f"{self.INFO['name']}-patchsize{self.patch_size}-len{len(self)}-{self.subset}"

    @property
    def truth(self) -> coo_array:
        '''alias for depreacted property'''
        return self.lbl
    


# SPDX-FileCopyrightText: 2025-present songyz2023 <songyz2023dlut@outlook.com>
# SPDX-License-Identifier: Apache-2.0

import unittest

import numpy as np
import skimage
from rs_fusion_datasets import fetch_augsburg_ouc, fetch_berlin_ouc, fetch_houston2018_ouc
from rs_fusion_datasets import fetch_houston2013, fetch_muufl, split_spmatrix, fetch_trento, Muufl, Houston2013, Houston2018Ouc, Trento, DataMetaInfo, lbl2rgb, AugsburgOuc, BerlinOuc, Houston2013Mmr
import torch
from torch.utils.data import DataLoader
from itertools import product
from hsi2rgb import hsi2rgb
from rs_fusion_datasets import * # test __all__ in __init__.py

from typing import get_type_hints


def is_typeddict_instance(obj, typeddict_cls):
    if not isinstance(obj, dict):
        return False
    type_hints = get_type_hints(typeddict_cls)
    for key, expected_type in type_hints.items():
        if key not in obj: # do not check value: isinstance() argument 2 cannot be a parameterized generic
            print(f"Key '{key}' is missing or has incorrect type.")
            return False
    return True

    

class Test(unittest.TestCase):
    def setUp(self):
        import logging
        # logging.basicConfig(level=logging.DEBUG)
        self.trento = fetch_trento()
        self.houston2013 = fetch_houston2013()
        self.muufl = fetch_muufl()
    
    def test_lbl2rgb(self):
        for datafetch in [fetch_augsburg_ouc, fetch_berlin_ouc, fetch_houston2018_ouc, fetch_houston2013, fetch_muufl, fetch_trento]:
            r = datafetch()
            if len(r) == 6:
                hsi, dsm, train_label, test_label, all_label,info = r
            elif len(r) == 5:
                hsi, dsm, train_label, test_label ,info = r
            elif len(r) == 4:
                hsi, dsm, label, info = r
                train_label, test_label = split_spmatrix(label, 100)
            else:
                raise ValueError(f"fetch function {datafetch} should return 4 or 5 or 6 elements, but got {len(r)}")
            self.generate_lbl2rgb(train_label, info, split='train')
            self.generate_lbl2rgb(test_label, info, split='test')



    def torch_dataloader_test(self, dataset):
        rgb = hsi2rgb(dataset.hsi, wavelength=dataset.INFO['wavelength'], input_format='CHW', output_format='HWC')
        skimage.io.imsave(f"dist/torch_{dataset.INFO['name']}_hsi.png", (rgb * 255.0).numpy().astype(np.uint8))
        for i,dsm in enumerate(dataset.dsm):
            dsm *= 255.0
            skimage.io.imsave(f"dist/torch_{dataset.INFO['name']}_dsm{i}.png", dsm.numpy().astype(np.uint8))

        b = 8
        dataloader = DataLoader(dataset, batch_size=b, shuffle=True, drop_last=True)
        x_h, x_l, y, extras = next(iter(dataloader))
        
        n_test = 10
        for x_h, x_l, y, extras in dataloader:
            if torch.cuda.is_available():
                x_h, x_l, y = x_h.cuda(), x_l.cuda(), y.cuda()
            self.assertIsInstance(x_h, torch.Tensor)
            self.assertIsInstance(x_l, torch.Tensor)
            self.assertIsInstance(y, torch.Tensor)
            self.assertEqual(x_h.shape, torch.Size([b, dataset.INFO['n_channel_hsi'],dataset.patch_size,dataset.patch_size]))
            self.assertEqual(x_l.shape, torch.Size([b, dataset.INFO['n_channel_dsm'],dataset.patch_size,dataset.patch_size]))
            self.assertEqual(y.shape, torch.Size([b, dataset.n_class]))
            self.assertEqual(x_h.dtype, torch.float)
            self.assertEqual(x_l.dtype, torch.float)
            self.assertEqual(y.dtype, torch.float)
            if n_test <= 0:
                break
            else:
                n_test -= 1

    def generate_lbl2rgb(self, truth, info, split):
        h,w = truth.shape
        y = np.eye(info['n_class']+1)[truth.todense()].transpose(2, 0, 1) # One Hot
        self.assertEqual(y.shape, (info['n_class']+1, h, w))

        rgb = lbl2rgb(y, info['name'])
        self.assertEqual(rgb.shape, (3, h, w))
        self.assertLessEqual(rgb.max(), 1.0)
        self.assertGreaterEqual(rgb.min(), 0.0)

        img = (rgb*255.0).astype(np.uint8).transpose(1, 2, 0)
        skimage.io.imsave(f"dist/{info['name']}_{split}.png", img, check_contrast=False)

    def test_fetch_houston2013(self):
        casi, lidar, train_truth, test_truth, info = self.houston2013
        H, W = 349, 1905
        C_H, C_L = 144, 1
        self.assertEqual(train_truth.data.max(), info['n_class'])
        self.assertEqual(train_truth.data.min(), 1)
        self.assertEqual(train_truth.todense().min(), 0)
        self.assertEqual(test_truth.data.max(), info['n_class'])
        self.assertEqual(test_truth.data.min(), 1)
        self.assertEqual(test_truth.todense().min(), 0)
        self.assertEqual(casi.shape, (C_H, H, W))
        self.assertEqual(lidar.shape, (C_L, H, W))
        self.assertEqual(train_truth.shape, (H, W))
        self.assertEqual(test_truth.shape, (H, W))
        self.assertEqual(info['n_channel_hsi'], 144)
        self.assertEqual(info['n_channel_dsm'], 1)
        self.assertEqual(info['n_class'], 15)
        self.assertEqual(info['width'], W)
        self.assertEqual(info['height'], H)
        self.assertEqual(len(info['label_name']), 15)
        self.assertEqual(info['wavelength'].shape, (144,))
        self.assertEqual(info['wavelength'][0], 364.000000)
        self.assertEqual(info['wavelength'][-1], 1046.099976)
        self.assertEqual(len(info['wavelength']), C_H)
        self.assertTrue(is_typeddict_instance(info, DataMetaInfo))



    def test_fetch_muufl(self):
        casi, lidar, truth, info = self.muufl
        train_label, test_label = split_spmatrix(truth, 20)
        H, W = 325, 220
        C_H, C_L = 64, 2
        self.assertEqual(casi.shape, (C_H, H, W))
        self.assertEqual(lidar.shape, (C_L, H, W))
        self.assertEqual(truth.shape, (H, W))
        self.assertEqual(train_label.shape, (H, W))
        self.assertEqual(test_label.shape, (H, W))
        self.assertEqual(info['n_channel_dsm'], C_L)
        self.assertEqual(info['n_channel_hsi'], C_H)
        self.assertEqual(len(info['wavelength']), C_H)
        self.assertTrue(is_typeddict_instance(info, DataMetaInfo))

    def test_fetch_trento(self):
        casi, lidar, truth, info = self.trento
        train_label, test_label = split_spmatrix(truth, 20)

        H, W = 166, 600
        C_H, C_L = 63, 2
        self.assertEqual(casi.shape, (C_H, H, W))
        self.assertEqual(lidar.shape, (C_L, H, W))
        self.assertEqual(info['n_channel_hsi'], C_H)
        self.assertEqual(len(info['wavelength']), C_H)
        self.assertEqual(info['n_channel_dsm'], C_L)

        self.assertEqual(truth.shape, (H, W))
        self.assertEqual(train_label.shape, (H, W))
        self.assertEqual(test_label.shape, (H, W))
        self.assertTrue(is_typeddict_instance(info, DataMetaInfo))

    def test_torch_datasets(self):
        for Dataset, split, patch_size in product([Houston2018Ouc, BerlinOuc,AugsburgOuc,Houston2013, Muufl, Trento, Houston2013Mmr], ['train', 'test', 'full'], [1, 6, 9]):
            dataset = Dataset(split=split, patch_size=patch_size)
            self.torch_dataloader_test(dataset)
            if Dataset in [Muufl, Trento] and split == 'train':
                n_train_perclass = 50
                self.assertEqual(dataset.n_class*n_train_perclass, len(Dataset(split=split, patch_size=5, n_train_perclass=n_train_perclass)))

    def test_datahome(self):
        fetch_trento(data_home='./tmp/')
        Trento(split='train', patch_size=5, root='./tmp/')
                    

if __name__ == '__main__':
    unittest.main()


from pathlib import Path
from typing import Optional, Union, Literal
import warnings

from ..core.fetch_houston2013 import fetch_houston2013
from ..core.fetch_houston2013_mmr import fetch_houston2013_mmr
from ..core.fetch_muufl import fetch_muufl
from ..core.fetch_trento import fetch_trento
from ..core.fetch_augsburg_ouc import fetch_augsburg_ouc
from ..core.fetch_berlin_ouc import fetch_berlin_ouc
from ..core.fetch_houston2018_ouc import fetch_houston2018_ouc
from ..core._fetch_muufl_mat import _fetch_muufl_mat
from ..core._fetch_trento_mat import _fetch_trento_mat
from ..core._fetch_houston2013_mat import _fetch_houston2013_mat
from ..util.split_spmatrix import split_spmatrix
from .common_hsi_dsm_dataset import CommonHsiDsmDataset

class Houston2018Ouc(CommonHsiDsmDataset):
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        """
        A preprocessed torch dataset for Houston 2018 (ouc) dataset.

        :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
        :param patch_size: The size of patches. Default is 5.
        :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
        """
        hsi, dsm, lbl_train, lbl_test, lbl_all,info = fetch_houston2018_ouc(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)


class BerlinOuc(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for Berlin (ouc) dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, lbl_all,info = fetch_berlin_ouc(data_home=root)
        dsm = dsm[::-1, :, :] # The first 3 channels are mostly black, so put the last channel to the first channel; TODO: check if the 3 dark channel are normal map or something important
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class AugsburgOuc(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the Augsburg 2018 (ouc) dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, lbl_all,info = fetch_augsburg_ouc(data_home=root)
        dsm = dsm[::-1, :, :] # The first 3 channels are mostly black, so put the last channel to the first channel; TODO: check if the 3 dark channel are normal map or something important
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class Houston2013(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the official Houston 2013 dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, info = fetch_houston2013(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class Houston2013Mmr(CommonHsiDsmDataset):
    """This is only for internal test."""
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, info = fetch_houston2013_mmr(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class _Houston2013Mat(CommonHsiDsmDataset):
    """This is only for internal test."""
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, info = _fetch_houston2013_mat(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class _Houston2013Mmrs(Houston2013Mmr):
    def __init__(self, *args, **kwargs):
        warnings.warn("_Houston2013Mmrs is deprecated, please use Houston2013Mmr instead.", DeprecationWarning)
        super().__init__(*args, **kwargs)

class Muufl(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the official Muufl dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param n_train_perclass: The number of training samples per class. Default is 100.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, n_train_perclass:Union[int, float]=100, *args, **kwargs):
        hsi, dsm, lbl, info = fetch_muufl(data_home=root)
        lbl_train, lbl_test = split_spmatrix(lbl, n_train_perclass)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)


class _MuuflMat(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the official Muufl dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param n_train_perclass: The number of training samples per class. Default is 100.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, n_train_perclass:Union[int, float]=100, *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, info = _fetch_muufl_mat(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)        

class Trento(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the Trento dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param n_train_perclass: The number of training samples per class. Default is 100.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None, n_train_perclass:Union[int, float]=100, *args, **kwargs):
        hsi, dsm, lbl, info = fetch_trento(data_home=root)
        lbl_train, lbl_test = split_spmatrix(lbl, n_train_perclass)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)

class _TrentoMat(CommonHsiDsmDataset):
    """
    A preprocessed torch dataset for the Trento dataset.

    :param split: 'train', 'test', 'full'. 'full' means the whole map, usually used for visualization.
    :param patch_size: The size of patches. Default is 5.
    :param root: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    """
    def __init__(self, split: Literal['train', 'test', 'full'], patch_size=5, root :Optional[Union[Path,str]]=None,  *args, **kwargs):
        hsi, dsm, lbl_train, lbl_test, info = _fetch_trento_mat(data_home=root)
        super().__init__(hsi, dsm, lbl_train, lbl_test, info, split, patch_size, *args, **kwargs)



from typing import TypedDict
import numpy as np

class DataMetaInfo(TypedDict):
    """
    DataMetaInfo is a TypedDict that defines the structure for storing metadata 
    and configuration information.

    Attributes:
        name (str): The short name (aka. dataset string id) of the dataset.
        full_name (str): The full descriptive name of the dataset.
        homepage (str): The URL of the dataset's homepage or source.
        n_channel_hsi (int): The number of hyperspectral channels in the dataset.
        n_channel_lidar (int): The number of LiDAR channels in the dataset.
        n_class (int): The number of classes in the dataset.
        width (int): The width of the dataset's images or data grid.
        height (int): The height of the dataset's images or data grid.
        label_dict (dict[int, str]): A dictionary mapping class indices to class names.
        wavelength (np.ndarray): An array containing the wavelengths of the hyperspectral bands.
    """
    name: str
    full_name: str
    homepage: str
    n_channel_hsi: int
    n_channel_dsm: int
    width: int
    height: int
    n_class: int
    label_name: dict[int, str]
    wavelength: np.ndarray


__all__ = [
    'DataMetaInfo',
]


import numpy as np
from torch import Tensor
from torch import argmax
from .lbl2rgb import lbl2rgb


class ClassificationMapper:
    def __init__(self, truth, dataset_name=None):
        self.dataset_name = dataset_name
        self.TRUTH = truth.todense()
        self.predict = np.zeros_like(self.TRUTH)

    def error(self):
        raise NotImplementedError("error() is not implemented yet")
        return self.predict != self.TRUTH

    def correct(self):
        raise NotImplementedError("error() is not implemented yet")
        return np.logical_and(self.predict == self.TRUTH, self.TRUTH != 0)

    def predict_image(self, format='chw'):
        img = lbl2rgb(self.predict, self.dataset_name)
        img = (img*255.0).astype(np.uint8)
        if format == 'hwc':
            return img.transpose(1, 2, 0)
        elif format == 'chw':
            return img
        else:
            raise ValueError(f"Unknown format: {format}")

    def add_sample(self, location, class_id, de_onehot=True):
        if de_onehot and isinstance(class_id, Tensor) and len(class_id.shape) > 1:
            class_id = argmax(class_id, dim=-1)
        self.predict[location[0].int().cpu(), location[1].int().cpu()] = class_id.cpu() + 1

    def __call__(self, location, class_id, de_onehot=True):
        self.add_sample(location, class_id, de_onehot)

    # def overlay_correct_image(self, underlying=None):
    #     return skimage.color.label2rgb(self.correct_image() + 2*self.error_image(), colors=['green', 'red'], alpha=0.5, bg_label=0, image=underlying)



from typing import List, Union
import numpy as np
import scipy.io

from scipy.sparse import coo_array
from jaxtyping import Float

from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo


def fetch_augsburg(url :Union[str, List[str]]='ftp://m1657312@dataserv.ub.tum.de?m1657312/Augsburg_data_4_publication.zip'):
    raise NotImplementedError("This dataset is implemented ")
    """
    Donwload and load the augsburg dataset.

    Use CHW format
    """

    basedir = zip_download_and_extract('augsburg', url, {
        'augsburg.zip'      :'b203331b039d994015c4137753f15973cb638046532b8dced6064888bf970631',
        'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/allgrd.mat'      :'7e3fb2a2ea22c2661dfc768db3cb93c9643b324e7e64fadedfa57f5edbf1818f',
        'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/Italy_hsi.mat'   :'7b965fd405314b5c91451042e547a1923be6f5a38c6da83969032cff79729280',
        'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/Italy_lidar.mat' :'a04dc90368d6a7b4f9d3936024ba9fef4105456c090daa14fff31b8b79e94ab1',
    })


    # 3. 数据加载
    hsi = scipy.io.loadmat(
        basedir / 'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/Italy_hsi.mat',
        squeeze_me=True,
        mat_dtype=True,
        struct_as_record=False
    )['data']

    lidar = scipy.io.loadmat(
        basedir / 'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/Italy_lidar.mat',
        squeeze_me=True,
        mat_dtype=True,
        struct_as_record=False
    )['data']

    truth = scipy.io.loadmat(
        basedir / 'augsburg-b4afc449ce5d6936ddc04fe267d86f9f35536afd/allgrd.mat',
        squeeze_me=True,
        mat_dtype=True,
        struct_as_record=False
    )['mask_test']
    truth = coo_array(truth)

    info :DataMetaInfo = {
        'name': 'augsburg',
        'full_name': 'augsburg',
        'version': '0.1',
        'homepage': '',
        'license': '',
        'n_channel_hsi': hsi.shape[-1],
        'n_channel_lidar': lidar.shape[-1],
        'n_class': 6,
        'width': hsi.shape[1],
        'height': hsi.shape[0],
        'label_name': {
            1: "Apple Trees",
            2: "Building",
            3: "Ground",
            4: "Woods",
            5: "Vineyard",
            6: "Roads"
        },
        'wavelength': np.linspace(402.89, 989.09, hsi.shape[-1])
    }

    return hsi.transpose(2,0,1), lidar.transpose(2,0,1), truth, info



__all__ = ['fetch_augsburg']



from pathlib import Path
from typing import List, Union, Optional, Tuple

from numpy import ndarray
import scipy.io

from scipy.sparse import coo_array
from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo
from jaxtyping import  UInt16, Float32, UInt8, Float64



def _fetch_muufl_mat(
        url :Union[str, List[str]]='http://10.7.36.2:5000/dataset/muufl_mat.zip', 
        data_home:Optional[Union[Path, str]]=None
        ) -> Tuple[
        Float32[ndarray, '64 325 220'], 
        Float32[ndarray, '2 325 220'],
        Float64[coo_array, '325 220'],
        DataMetaInfo
    ]:
    """Fetch and load the Muufl dataset.

    The background is 0, and the labels start from 1. All images are CHW formats.
    
    :param url: The URL to download the dataset. Use a list to specify multiple mirrored URLs.
    :param data_home: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    :return: (hsi, lidar, train_labels, train_labels, info)
    """

    basedir = zip_download_and_extract('muufl-mat', url, {
        'muufl-mat.zip'       : '95cd1235a5f02df69cd384de5fb83d0d14fa37ac5776081caf9d477368e9a291',
        'muufl_mat/HSI.mat'   : 'adf5934333753b2e05cf244392b43d86b432aa8dac039964ce1f409c96c65edd',
        'muufl_mat/labels.mat': 'a114cec5f9b51c91fac3041426a7142e6dbc0e6d895af9b3b5cf7819e7c415b0',
        'muufl_mat/Lidar.mat' : '31a9f7d6b2e9ab6507304c72b7b99ec975565fd5961ea9523827443690bb7194'
    }, data_home=data_home)

    hsi :Float32[ndarray, '325 220 64'] = scipy.io.loadmat(
        basedir / 'muufl_mat/HSI.mat',
        squeeze_me=True,
        mat_dtype=True,
        struct_as_record=False
    )['HSI']
    lidar :Float32[ndarray, '325 220 2'] = scipy.io.loadmat(
        basedir / 'muufl_mat/Lidar.mat',
        squeeze_me=True,
        mat_dtype=True,
        struct_as_record=False
    )['Lidar']
    truth :Float64[ndarray, '325 220'] = d.sceneLabels.labels
    truth[truth==-1] = 0
    truth = coo_array(truth, dtype='int')
    info :DataMetaInfo = {
        'name': 'muufl-mat',
        'full_name': 'MUUFL Gulfport dataset, a mat file that can not find where it comes, only for internal use',
        'version': '0.1',
        'homepage': 'https://github.com/GatorSense/MUUFLGulfport',
        'license': 'MIT',
        'n_channel_hsi': hsi.shape[-1],
        'n_channel_dsm': lidar.shape[-1],
        'n_class': d.sceneLabels.Materials_Type.size,
        'width': hsi.shape[1],
        'height': hsi.shape[0],
        'label_name': dict(enumerate(d.sceneLabels.Materials_Type, start=1)),
        'wavelength': d.info.wavelength
    }

    return hsi.transpose(2,0,1), lidar.transpose(2,0,1), truth, info




__all__ = ['fetch_muufl', 'split_spmatrix']



# SPDX-FileCopyrightText: 2025-present songyz2023 <songyz2023dlut@outlook.com>
# SPDX-License-Identifier: Apache-2.0

from typing import List, Union

import numpy as np
import rasterio
from scipy.sparse import coo_array, spmatrix
from jaxtyping import UInt16, Float32, UInt64

from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo

def fetch_houston2018(url :Union[str, List[str]]='http://machinelearning.ee.uh.edu/QZ23es1aMPH/2018IEEE/phase2.zip') -> tuple[
    UInt16[np.ndarray, '144 349 1905'],
    Float32[np.ndarray, '1 349 1905'],
    UInt64[spmatrix, '349 1905'],
    UInt64[spmatrix, '349 1905'],
    DataMetaInfo
]:
    """Load the Houston2018 data-set in scikit-learn style

    Download it if necessary. All the image are CHW formats. And the shape are typed in the return type.
    The background label is 0, and the rest are 1-15.
    
    :param datahome: The path to store the data files, default is '~/scikit_learn_data'
    :param download_if_missing: Whether to download the data if it is not found

    :return: (hsi, dsm, train_truth, test_truth, info)
    """
    basedir=zip_download_and_extract('houston2018', url, {
        'houston2018.zip': 'aa6ae39bf9df16b6a72eddaec7e05d853ef525a2f43555f1b9394ec0fe801f19',
        "2018IEEE_Contest/Phase2/TrainingGT/2018_IEEE_GRSS_DFC_GT_TR.tif": "025b3813a0a52855567debf4b1e294937fd9b1ff17e00b9acd9e794ef29696d0",
        "2018IEEE_Contest/Phase2/FullHSIDataset/20170218_UH_CASI_S4_NAD83.pix": "6f29fd1863ceb0cf4ae3b162ffc48e11ce02bad46780531a99ac2b29164b38e2",
        '2018IEEE_Contest/Phase2/Lidar GeoTiff Rasters/DEM+B_C123/UH17_GEM051.tif': '521b04abe70ef1452cdd74ee46b30efa4cc093e707a365bbdd6ad54f6317779a'
    })

    # 3. 数据加载
    with rasterio.open(basedir / '2018IEEE_Contest/Phase2/FullHSIDataset/20170218_UH_CASI_S4_NAD83.pix') as f:
        hsi = f.read() # (50, 1202, 4172)
    with rasterio.open(basedir / '2018IEEE_Contest/Phase2/Lidar GeoTiff Rasters/DEM+B_C123/UH17_GEM051.tif') as f:
        dem = f.read() # (1, 2404, 8344)
    with rasterio.open(basedir / '2018IEEE_Contest/Phase2/TrainingGT/2018_IEEE_GRSS_DFC_GT_TR.tif') as f:
        lbl = f.read() # (1,1202,4768) TODO shape are different
        lbl = coo_array(lbl, dtype='int')

    info :DataMetaInfo = {
        'name': 'houston2018',
        'full_name': 'IEEE GRSS DF Contest Houston 2018',
        'homepage': 'https://machinelearning.ee.uh.edu/2018-ieee-grss-data-fusion-challenge-fusion-of-multispectral-lidar-and-hyperspectral-data/',
        'n_channel_hsi': 50,
        'n_channel_lidar': 1,
        'n_class': 20,
        'width': -1,
        'height': -1,
        'label_name': {
            0: 'Unclassified',
            1: 'Healthy grass',
            2: 'Stressed grass',
            3: 'Artificial turf',
            4: 'Evergreen trees',
            5: 'Deciduous trees',
            6: 'Bare earth',
            7: 'Water',
            8: 'Residential buildings',
            9: 'Non-residential buildings',
            10: 'Roads',
            11: 'Sidewalks',
            12: 'Crosswalks',
            13: 'Major thoroughfares',
            14: 'Highways',
            15: 'Railways',
            16: 'Paved parking lots',
            17: 'Unpaved parking lots',
            18: 'Cars',
            19: 'Trains',
            20: 'Stadium seats',
        },
        "wavelength": np.array([
            374.399994,  388.700012,  403.100006,  417.399994,  431.700012,  446.100006,
            460.399994,  474.700012,  489.000000,  503.399994,  517.700012,  532.000000,
            546.299988,  560.599976,  574.900024,  589.200012,  603.599976,  617.900024,
            632.200012,  646.500000,  660.799988,  675.099976,  689.400024,  703.700012,
            718.000000,  732.299988,  746.599976,  760.900024,  775.200012,  789.500000,
            803.799988,  818.099976,  832.400024,  846.700012,  861.099976,  875.400024,
            889.700012,  904.000000,  918.299988,  932.599976,  946.900024,  961.200012,
            975.500000,  989.799988, 1004.200012, 1018.500000, 1032.800049, 1047.099976,
            49.000000,   50.000000
        ])
    }

    return hsi, dem, lbl, info


__all__ = ['fetch_houston2018']





from typing import List, Tuple, Union, Optional
from pathlib import Path
import numpy as np
from numpy import ndarray
import scipy.io

from scipy.sparse import coo_array
from jaxtyping import  UInt16, Float32, UInt8, Int8, Int4

from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo


def fetch_houston2018_ouc(
        url       :Union[str, List[str]]      = 'https://github.com/songyz2019/rs-fusion-datasets-dist/releases/download/v1.0.0/houston2018-ouc.zip', 
        data_home :Optional[Union[Path, str]] = None
    ) -> Tuple[
        UInt16[ndarray, '50 1202 4768'], 
        Float32[ndarray, '1 1202 4768'],
        UInt8[coo_array, '1202 4768'],
        UInt8[coo_array, '1202 4768'],
        UInt8[coo_array, '1202 4768'],
        DataMetaInfo
    ]:
    """Fetch and load the Houston2018 (ouc) dataset. 

    The background is 0, and the labels start from 1. All images are CHW formats.
    
    :param url: The URL to download the dataset. Use a list to specify multiple mirrored URLs.
    :param data_home: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    :return: (hsi, lidar, train_labels, train_labels, info)
    """

    basedir = zip_download_and_extract('houston2018-ouc', url, {
        'houston2018-ouc.zip'          :'bc1e046cc00b82661441888b17327b38946a0da33951eb4ebb7d0a4e0ca3cfa2',
        'Houston2018/houston_gt.mat'   : '93f268e576fcf149fab8cb71b4a79691b420e9a3f369f0cbc73e2029639f6286',
        'Houston2018/houston_hsi.mat'  : '35a4a9a2bad193740274c6388f18284fc586a2ef5bde579368796fbeec2a58ea',
        'Houston2018/houston_index.mat': '2d513db197f105b6b82e7039909878862fbe48750a93184017f22ab8f288d0b3',
        'Houston2018/houston_lidar.mat': 'a55fa2d99684078dd0d8039b2ca1349190d86766426a25e8dff7fbd5df036d9b',
    }, data_home=data_home)


    # 3. 数据加载
    hsi   :UInt16[ndarray, '1202 4768 50']= scipy.io.loadmat(
        basedir / 'Houston2018/houston_hsi.mat',
    )['houston_hsi']

    lidar :Float32[ndarray, '1202 4768']   = scipy.io.loadmat(
        basedir / 'Houston2018/houston_lidar.mat',
    )['houston_lidar']
    lidar = lidar[:,:,np.newaxis]

    lbl_map :UInt8[ndarray, '1202 4768']   = scipy.io.loadmat(
        basedir / 'Houston2018/houston_gt.mat',
    )['houston_gt']
    lbl_subsets = scipy.io.loadmat(basedir / 'Houston2018/houston_index.mat')
    lbl_train_coord :Int8[ndarray, 'M 2']    = lbl_subsets['houston_train']   # TODO: replace M N Q with the actual number
    lbl_test_coord  :Int8[ndarray, 'N 2']  = lbl_subsets['houston_test']
    lbl_all_coord   :Int4[ndarray, 'Q 2'] = lbl_subsets['houston_all']
    lbl_train :UInt8[coo_array, '1202 4768'] = coo_array((
        lbl_map[ lbl_train_coord[:,0],lbl_train_coord[:,1] ],
        lbl_train_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_test :UInt8[coo_array, '1202 4768']  = coo_array((
        lbl_map[ lbl_test_coord[:,0],lbl_test_coord[:,1] ],
        lbl_test_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_all  :UInt8[coo_array, '1202 4768']  = coo_array((
        lbl_map[ lbl_all_coord[:,0],lbl_all_coord[:,1] ],
        lbl_all_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    info :DataMetaInfo = {
        'name': 'houston2018-ouc',
        'full_name': 'preprocessed houston2018 dataset',
        'version': '',
        'homepage': 'https://github.com/oucailab/DCMNet/',
        'license': '',
        'n_channel_hsi': 50,
        'n_channel_dsm': 1,
        'n_class': 20,
        'width': 4768,
        'height': 1202,
        'label_name': {
            1: 'Healthy grass',
            2: 'Stressed grass',
            3: 'Artificial turf',
            4: 'Evergreen trees',
            5: 'Deciduous trees',
            6: 'Bare earth',
            7: 'Water',
            8: 'Residential buildings',
            9: 'Non-residential buildings',
            10: 'Roads',
            11: 'Sidewalks',
            12: 'Crosswalks',
            13: 'Major thoroughfares',
            14: 'Highways',
            15: 'Railways',
            16: 'Paved parking lots',
            17: 'Unpaved parking lots',
            18: 'Cars',
            19: 'Trains',
            20: 'Stadium seats',
        },
        "wavelength": np.array([
            374.399994,  388.700012,  403.100006,  417.399994,  431.700012,  446.100006,
            460.399994,  474.700012,  489.000000,  503.399994,  517.700012,  532.000000,
            546.299988,  560.599976,  574.900024,  589.200012,  603.599976,  617.900024,
            632.200012,  646.500000,  660.799988,  675.099976,  689.400024,  703.700012,
            718.000000,  732.299988,  746.599976,  760.900024,  775.200012,  789.500000,
            803.799988,  818.099976,  832.400024,  846.700012,  861.099976,  875.400024,
            889.700012,  904.000000,  918.299988,  932.599976,  946.900024,  961.200012,
            975.500000,  989.799988, 1004.200012, 1018.500000, 1032.800049, 1047.099976,
            49.000000,   50.000000
        ])
    }

    return hsi.transpose(2,0,1), lidar.transpose(2,0,1), lbl_train, lbl_test, lbl_all, info



__all__ = ['fetch_trento']



# understand the code of joint classification of HSI and DSM data in about 50 lines
# if you're using v0.12.3 and before, you should remove ClassificationMapper and its related code to make it work
import skimage
import torch
from torch import argmax
from torch.nn import Sequential, LazyConv2d, ReLU, LazyBatchNorm2d, Module, LazyLinear
from torch.nn.functional import cross_entropy
from torch.optim import Adam
from torch.utils.data import DataLoader
from rs_fusion_datasets import Houston2013, Trento, Muufl, ClassificationMapper

class MyModel(Module):
    def __init__(self, n_class):
        super().__init__()
        self.conv_hsi = Sequential(
            LazyConv2d(64, kernel_size=3), ReLU(), LazyBatchNorm2d(),
            LazyConv2d(16, kernel_size=3)
        )
        self.conv_dsm = Sequential(
            LazyConv2d(8,  kernel_size=3), ReLU(), LazyBatchNorm2d(),
            LazyConv2d(16, kernel_size=3)
        )
        self.classifier = LazyLinear(n_class)
    
    def forward(self, hsi, dsm):
        x = self.conv_hsi(hsi) + self.conv_dsm(dsm)
        return self.classifier(x.flatten(start_dim=1))
if __name__=='__main__':
    # Train
    trainset = Houston2013('train', patch_size=5)
    model = MyModel(n_class=trainset.INFO['n_class'])
    optimizer = Adam(model.parameters(), lr=0.005)
    for epoch in range(10):
        for hsi,dsm,lbl,info in DataLoader(trainset, batch_size=64, shuffle=True, drop_last=True):
            optimizer.zero_grad()
            y_hat = model(hsi, dsm)
            loss = cross_entropy(y_hat, lbl)
            loss.backward()
            optimizer.step()
            print(f"{epoch=} {loss=}")
    torch.save(model.state_dict(), 'model.pth')
    # Test
    testset = Houston2013('test', patch_size=5)
    mappper = ClassificationMapper(testset.lbl, dataset_name=testset.INFO['name'])
    n_correct = 0
    model.eval()
    for hsi,dsm,lbl,info in DataLoader(testset, batch_size=1):
        y_hat = model(hsi, dsm)
        if argmax(y_hat, dim=1) == argmax(lbl, dim=1):
            n_correct += 1
            mappper.add_sample(info['location'], y_hat)
            print(f"accuracy: {n_correct}/{len(testset)}")
    skimage.io.imsave('result.png', mappper.predict_image(format='hwc'))


from colorsys import hsv_to_rgb
from typing import Union
import skimage
import torch
import numpy as np
from jaxtyping import Float

def hue_platte(n_sample :int):
    return [hsv_to_rgb(n/n_sample, 1, 1) for n in range(n_sample)]

def lbl2rgb(lbl :Float[Union[np.ndarray,torch.Tensor], '... C H W'], palette, kind='overlay') -> Float[Union[np.ndarray,torch.Tensor], '... 3 H W']:
    """
    符合实验室内部 格式要求的OneHot标签转图像函数

    :param kind: 分割模式,如果为'overlay'则是硬分割, 'avg'则是软分割
    :param lbl B C H W格式的图像batch, C为OneHot编码格式
    :return: B 3 H W格式的RGB图像, 取值范围为0~1
    """
    def hex2rgb(x :str):
        if not x.startswith("#"):
            return x
        x = x.removeprefix("#")
        r = int(x[:2],  base=16) / 255.0
        g = int(x[2:4], base=16) / 255.0
        b = int(x[4:6], base=16) / 255.0
        return [r,g,b]
    placeholder_palette = [
        "#e6194b", "#3cb44b", "#ffe119", "#4363d8", "#f58231", "#911eb4", "#46f0f0", "#f032e6",
        "#bcf60c", "#fabebe", "#008080", "#e6beff", "#9a6324", "#fffac8", "#800000", "#aaffc3",
        "#808000", "#ffd8b1", "#000075", "#808080", "#ffffff", "#000000", "#1f77b4", "#ff7f0e",
        "#2ca02c", "#d62728", "#9467bd", "#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"
    ]
    palette_presets = {
        'houston2013': ('forestgreen', 'limegreen', 'darkgreen', 'green', 'indianred', 'royalblue', 'papayawhip', 'pink','red', 'orangered', 'cadetblue', 'yellow', 'darkorange', 'darkmagenta', 'cyan'),
        'houston2013-mmr': ('forestgreen', 'limegreen', 'darkgreen', 'green', 'indianred', 'royalblue', 'papayawhip', 'pink','red', 'orangered', 'cadetblue', 'yellow', 'darkorange', 'darkmagenta', 'cyan'),
        'houston2013-mat': ('forestgreen', 'limegreen', 'darkgreen', 'green', 'indianred', 'royalblue', 'papayawhip', 'pink','red', 'orangered', 'cadetblue', 'yellow', 'darkorange', 'darkmagenta', 'cyan'),
        'muufl': ('forestgreen', 'limegreen', 'lightblue', 'papayawhip', 'red', 'blue', 'purple', 'pink','orangered', 'yellow', 'brown'),
        'muufl-mat': ('forestgreen', 'limegreen', 'lightblue', 'papayawhip', 'red', 'blue', 'purple', 'pink','orangered', 'yellow', 'brown'),
        'trento': ('royalblue','lightblue' , 'limegreen', 'yellow', 'red', 'brown'),
        'trento-mat': ('royalblue','lightblue' , 'limegreen', 'yellow', 'red', 'brown'),
        'houston2018-ouc': placeholder_palette, #TODO
        'augsburg-ouc':    placeholder_palette, #TODO
        'berlin-ouc':      placeholder_palette, #TODO
    }
    if palette in palette_presets:
        palette = palette_presets[palette]
    elif not isinstance(palette, tuple):
        raise ValueError(f"palette should be a tuple of colors, or a string in {list(palette_presets.keys())}, but got {palette}")
    
    if len(lbl.shape)==3:
        if isinstance(lbl, torch.Tensor):
            lbl = torch.argmax(lbl, dim=-3).cpu().numpy()
        else:
            lbl = np.argmax(lbl, axis=-3)
    img = skimage.color.label2rgb(
        lbl,
        channel_axis=-3,
        colors=[hex2rgb(x) for x in palette],
        bg_label=0,
        bg_color=hex2rgb('#000000'),
        kind=kind # 硬分割还是软分割
    )
    return img


from typing import List, Tuple, Union, Optional
from pathlib import Path
import numpy as np
from numpy import ndarray
import scipy.io

from scipy.sparse import coo_array
from jaxtyping import  UInt16, Float64, UInt8, Int8, Int4

from ..util.fileio import zip_download_and_extract
from .common import DataMetaInfo


def fetch_augsburg_ouc(
        url       :Union[str, List[str]]      = 'https://github.com/songyz2019/rs-fusion-datasets-dist/releases/download/v1.0.0/augsburg-ouc.zip', 
        data_home :Optional[Union[Path, str]] = None
    ) -> Tuple[
        UInt16[ndarray, '180 332 485'], 
        Float64[ndarray, '4 332 485'],
        UInt8[coo_array, '332 485'],
        UInt8[coo_array, '332 485'],
        UInt8[coo_array, '332 485'],
        DataMetaInfo
    ]:
    """Fetch and load the Augsburg (ouc) dataset. 

    The background is 0, and the labels start from 1. All images are CHW formats.
    
    :param url: The URL to download the dataset. Use a list to specify multiple mirrored URLs.
    :param data_home: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    :return: (hsi, lidar, train_labels, train_labels, info)
    """

    basedir = zip_download_and_extract('augsburg-ouc', url, {
        'augsburg-ouc.zip'            :'4d853be609eca98a5fb23ccd711bcd0985b7db86a6e89f53f461e1643a7ab6ae',
        'Augsburg/augsburg_gt.mat'    : '065df71257f06c075ff3b37327c0fba75cc9146015a0f095a093fd12b9eaacbe',
        'Augsburg/augsburg_hsi.mat'   : '2f13839d7840b32f55e0489d6934776dcfded7f678e23a9d8c2c092573946583',
        'Augsburg/augsburg_index.mat' : '8bc57037447cb42dbf3a8478d42d4bc217a53a6a0b349c2d67db8c660bef2f43',
        'Augsburg/augsburg_sar.mat'   : 'cd1dbfad2346f47e51d105b05e11aaefc872fe487a94bdc2913f4eaee3f8e82b'
    }, data_home=data_home)


    # 3. 数据加载
    hsi   :UInt16[ndarray, '332 485 180']= scipy.io.loadmat(
        basedir / 'Augsburg/augsburg_hsi.mat',
    )['augsburg_hsi']

    sar :Float64[ndarray, '332 485 4']   = scipy.io.loadmat(
        basedir / 'Augsburg/augsburg_sar.mat',
    )['augsburg_sar']

    lbl_map :UInt8[ndarray, '332 485']   = scipy.io.loadmat(
        basedir / 'Augsburg/augsburg_gt.mat',
    )['augsburg_gt']
    lbl_subsets = scipy.io.loadmat(basedir / 'Augsburg/augsburg_index.mat')
    lbl_train_coord :Int8[ndarray, '761 2']    = lbl_subsets['augsburg_train'] 
    lbl_test_coord  :Int8[ndarray, '77533 2']  = lbl_subsets['augsburg_test']
    lbl_all_coord   :Int4[ndarray, '161020 2'] = lbl_subsets['augsburg_all']
    lbl_train :UInt8[coo_array, '332 485'] = coo_array((
        lbl_map[ lbl_train_coord[:,0],lbl_train_coord[:,1] ],
        lbl_train_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_test :UInt8[coo_array, '332 485']  = coo_array((
        lbl_map[ lbl_test_coord[:,0],lbl_test_coord[:,1] ],
        lbl_test_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)
    lbl_all  :UInt8[coo_array, '332 485']  = coo_array((
        lbl_map[ lbl_all_coord[:,0],lbl_all_coord[:,1] ],
        lbl_all_coord.transpose(1,0),
    ),dtype=lbl_map.dtype,shape=lbl_map.shape)

    info :DataMetaInfo = {
        'name': 'augsburg-ouc',
        'full_name': 'preprocessed augsburg dataset',
        'version': '',
        'homepage': 'https://github.com/oucailab/DCMNet/',
        'license': '',
        'n_channel_hsi': 180,
        'n_channel_dsm': 4,
        'n_class': 8,
        'width': 485,
        'height': 332,
        'label_name': {
            1: 'Forest',
            2: 'Residential',
            3: 'Industrial',
            4: 'Lowplants',
            5: 'Soil',
            6: 'Allotment',
            7: 'Commercial',
            8: 'Water',

        },
        'wavelength': np.linspace(410, 2500, 180) # TODO: check the wavelength
    }

    return hsi.transpose(2,0,1), sar.transpose(2,0,1), lbl_train, lbl_test, lbl_all, info



__all__ = ['fetch_trento']



from typing import List, Tuple, Union, Optional
from pathlib import Path
import numpy as np
from numpy import ndarray
import scipy.io

from scipy.sparse import coo_array

from ..util.fileio import zip_download_and_extract, load_one_key_mat
from .common import DataMetaInfo
from jaxtyping import  Float32, UInt8


def _fetch_trento_mat(
        url       :Union[str, List[str]]      = 'http://10.7.36.2:5000/dataset/trento_mat.zip', 
        data_home :Optional[Union[Path, str]] = None
        ) -> Tuple[
        Float32[ndarray, '63 166 600'], 
        Float32[ndarray, '2 166 600'],
        UInt8[coo_array, '166 600'],
        DataMetaInfo
    ]:
    """Fetch and load the Trento dataset.

    The background is 0, and the labels start from 1. All images are CHW formats.
    
    :param url: The URL to download the dataset. Use a list to specify multiple mirrored URLs.
    :param data_home: The path to store the data files, default is SCIKIT_LEARN_DATA environment variable or '~/scikit_learn_data'
    :return: (hsi, lidar, train_labels, train_labels, info)
    """

    basedir = zip_download_and_extract('trento-mat', url, {
        'trento-mat.zip'        : 'ae8a2e498a51ca0aad8995e86f1533390c880197a8a9cd96b7d28fe6e962274b',
        'Trento/HSI.mat'    : '44a84b4fe8bab9b2f5ec11af32602436ccc97653b7ae8c3283334da7d221c053',
        'Trento/LiDAR.mat'  : 'd99347532e7b18982c9537fd4de43574dd64f9b35ccab251459e7b483ddd43df',
        'Trento/TRLabel.mat': '593976a1bef4edf06ec543926811509f997454557aa9313ed6165b3012c26e07',
        'Trento/TSLabel.mat': '629ae33464907098cacf58549996711bf7079f036946f9208fde8a67842ab499'
    }, data_home=data_home)


    # 3. 数据加载
    hsi  :Float32[ndarray, '166 600 63']  = load_one_key_mat(
        basedir / 'Trento/HSI.mat',
    )

    lidar :Float32[ndarray, '166 600']  = load_one_key_mat(
        basedir / 'Trento/LiDAR.mat',
    )
    lidar = np.expand_dims(lidar, axis=-1)

    lbl_train :UInt8[ndarray, '166 600'] = load_one_key_mat(
        basedir / 'Trento/TRLabel.mat',
    )
    lbl_train = coo_array(lbl_train)

    lbl_test :UInt8[ndarray, '166 600'] = load_one_key_mat(
        basedir / 'Trento/TSLabel.mat',
    )
    lbl_test = coo_array(lbl_test)


    info :DataMetaInfo = {
        'name': 'trento-mat',
        'full_name': 'Trento',
        'version': '0.1',
        'homepage': '',
        'license': '',
        'n_channel_hsi': hsi.shape[-1],
        'n_channel_dsm': lidar.shape[-1],
        'n_class': 6,
        'width': hsi.shape[1],
        'height': hsi.shape[0],
        'label_name': {
            1: "Apple Trees",
            2: "Building",
            3: "Ground",
            4: "Woods",
            5: "Vineyard",
            6: "Roads"
        },
        'wavelength': np.linspace(402.89, 989.09, hsi.shape[-1])
    }

    return hsi.transpose(2,0,1), lidar.transpose(2,0,1), lbl_train, lbl_test, info



__all__ = ['fetch_trento']
